#include <mos/mos_os.h>
#include <mos/mos_fileio.h>

#include "phidget22.h"
#include "phidget22extra.h"

#define BUFSIZE (128 * 1024)

static const char *infile;
static const char *outfile;
static const char *dkey;
static const char *gkey;
static const char *skey;
static const char *val;

static int jflag;
static int vflag;
static int Bflag;
static int Sflag;
static int Nflag;
static int Iflag;
static int Vflag;

static void
usage(const char *argv0, int err) {
	FILE *out = err ? stderr : stdout;

	fprintf(out, "usage: %s -c <in> [-V][-j] [-o <out>] "
				 "[-d <key>][-g <key>][-s <key> -B<b>|-S<str>|-I<int>|-N<num>]\n",
			argv0);
	fprintf(out, " -c <in>          the pconf file to read\n");
	fprintf(out, " -V               enable verbose output (supports multiple)\n");
	fprintf(out, " -v <in>          the pconf file to validate\n");
	fprintf(out, " -d <key>         the key to delete\n");
	fprintf(out, " -g <key>         the key to get\n");
	fprintf(out, " -s <key>         the key to set\n");
	fprintf(out, " -B <b>           set boolean value\n");
	fprintf(out, " -I <int>         set integer value\n");
	fprintf(out, " -S <str>         set string value\n");
	fprintf(out, " -N <num>         set numeric value\n");
	fprintf(out, " -j               reads the input file as JSON\n");
	fprintf(out, " -o <out>         the file to write (converting from JSON to pc or pc to JSON)\n");
	fprintf(out, "                   -o is always for a conversion, and creates a new file\n");

	exit(err);
}

static PhidgetReturnCode
readx(char *buf, size_t bufsz, const char *file) {
	PhidgetReturnCode res;
	size_t n;
	int i;

	for (i = 0; i < 6; i++) {
		n = bufsz;
		res = mos_file_readx(MOS_IOP_IGNORE, buf, &n, "%s", file);
		if (res == 0) {
			buf[n] = '\0';
			return (EPHIDGET_OK);
		}
		if (res == MOSN_BUSY && i < 5) {
			suplogerr("'%s' busy: waiting", file);
			mos_usleep(250000);
			continue;
		}
		suplogerr("failed to read from '%s'", file);
		return (res);
	}
	return (EPHIDGET_IO);
}

static PhidgetReturnCode
writex(const char *buf, size_t bufsz, const char *file) {
	PhidgetReturnCode res;
	int i;

	for (i = 0; i < 6; i++) {
		res = mos_file_writex(MOS_IOP_IGNORE, buf, bufsz, "%s", file);
		if (res == 0)
			return (EPHIDGET_OK);

		if (res == MOSN_BUSY && i < 5) {
			suplogerr("'%s' busy: waiting", file);
			mos_usleep(250000);
			continue;
		}
		suplogerr("failed to write to '%s'", file);
		return (res);
	}
	return (EPHIDGET_IO);
}

int main(int argc, char **argv) {
	PhidgetReturnCode res;
	const char *argv0;
	char errbuf[128];
	pconf_t *pc;
	int64_t i64;
	double dbl;
	int32_t bl;
	char *buf;
	char *end;
	int ch;

	res = EPHIDGET_UNEXPECTED;
	argv0 = argv[0];
	pc = NULL;

	while ((ch = mos_getopt(argc, argv, "c:d:g:s:o:v:jB:I:N:S:V-")) != -1) {
		switch (ch) {
		case 'c':
			infile = mos_optarg;
			break;
		case 'v':
			vflag = 1;
			infile = mos_optarg;
			break;
		case 'd':
			dkey = mos_optarg;
			break;
		case 'g':
			gkey = mos_optarg;
			break;
		case 's':
			skey = mos_optarg;
			break;
		case 'o':
			outfile = mos_optarg;
			break;
		case 'j':
			jflag = 1;
			break;
		case 'B':
			Bflag = 1;
			val = mos_optarg;
			break;
		case 'I':
			Iflag = 1;
			val = mos_optarg;
			break;
		case 'N':
			Nflag = 1;
			val = mos_optarg;
			break;
		case 'S':
			Sflag = 1;
			val = mos_optarg;
			break;
		case 'V':
			Vflag++;
			break;
		case '-':
			usage(argv[0], 0); /* To handle --help and --version arguments */
							   /* NOT REACHED */
		default:
			usage(argv0, 1);
			/* NOT REACHED */
		}
	}
	argc -= mos_optind;
	argv += mos_optind;

	if (infile == NULL)
		usage(argv0, 1);

	if (skey && val == NULL)
		usage(argv0, 1);

	switch (Vflag) {
	case 0:
		PhidgetLog_enable(PHIDGET_LOG_ERROR, NULL);
		PhidgetLog_addSource(PSUPLS, PHIDGET_LOG_INFO);
		break;
	case 1:
		PhidgetLog_enable(PHIDGET_LOG_INFO, NULL);
		PhidgetLog_addSource(PSUPLS, PHIDGET_LOG_INFO);
		break;
	case 2:
		PhidgetLog_enable(PHIDGET_LOG_INFO, NULL);
		PhidgetLog_addSource(PSUPLS, PHIDGET_LOG_VERBOSE);
		break;
	case 3:
		PhidgetLog_enable(PHIDGET_LOG_INFO, NULL);
		PhidgetLog_addSource(PSUPLS, PHIDGET_LOG_VERBOSE);
		PhidgetLog_setSourceLevel("_pconf", PHIDGET_LOG_DEBUG);
		break;
	default:
	case 4:
		PhidgetLog_enable(PHIDGET_LOG_VERBOSE, NULL);
		PhidgetLog_addSource(PSUPLS, PHIDGET_LOG_VERBOSE);
		PhidgetLog_setSourceLevel("_pconf", PHIDGET_LOG_VERBOSE);
	}

	suplogdebug("infile:%s", infile);
	buf = mos_malloc(BUFSIZE);

	if (jflag) {
		suplogdebug("parsing JSON");
		res = readx(buf, BUFSIZE, infile);
		if (res != EPHIDGET_OK) {
			suplogerr("failed to read '%s'", infile);
			goto bad;
		}
		res = pconf_parsejson(&pc, buf, mos_strlen(buf));
		if (res != EPHIDGET_OK) {
			suplogerr("failed to parse JSON '%s'", infile);
			goto bad;
		}
	} else {
		suplogdebug("parsing pconf");
		res = pconf_parsepc_locked(&pc, errbuf, sizeof(errbuf), "%s", infile);
		if (res != EPHIDGET_OK) {
			suplogerr("failed to parse '%s' (%s)", infile, errbuf);
			goto bad;
		}
	}

	/*
	 * If verifying, we are done.
	 */
	if (vflag) {
		if (res == EPHIDGET_OK)
			printf("'%s' ok\n", infile);
		goto done;
	}

	pconf_setcreatemissing(pc, 1);

	if (dkey) {
		suplogdebug("removing %s", dkey);
		res = pconf_remove(pc, dkey);
		if (res != EPHIDGET_OK) {
			suplogerr("failed to remove key '%s': %s", dkey, getErrorStr(res));
			goto bad;
		}
	} else if (skey) {
		suplogdebug("setting %s", skey);
		if (pconf_exists(pc, "%s", skey)) {
			suplogdebug("  %s exists: removing", skey);
			res = pconf_remove(pc, "%s", skey);
			if (res != EPHIDGET_OK) {
				suplogerr("failed to update '%s'", skey);
				goto bad;
			}
		}
		if (Sflag) {
			suplogdebug("%s = (str)%s", skey, val);
			res = pconf_addstr(pc, val, "%s", skey);
			if (res != EPHIDGET_OK) {
				suplogerr("failed to set key '%s'", skey);
				goto bad;
			}
		} else if (Bflag) {
			suplogdebug("%s = (bool)%s", skey, val);
			res = mos_strto32(val, 0, &bl);
			if (res != 0) {
				if (mos_strcasecmp(val, "true") == 0)
					bl = 1;
				else
					bl = 0;
			}
			res = pconf_addbool(pc, bl, "%s", skey);
			if (res != EPHIDGET_OK) {
				suplogerr("failed to set key '%s'", skey);
				goto bad;
			}
		} else if (Nflag) {
			suplogdebug("%s = (dbl)%s", skey, val);
			dbl = strtod(val, &end);
			if (dbl == 0 && end == val) {
				suplogerr("invalid number: %s", val);
				goto bad;
			}
			res = pconf_addnum(pc, dbl, "%s", skey);
			if (res != EPHIDGET_OK) {
				suplogerr("failed to set key '%s'", skey);
				goto bad;
			}
		} else if (Iflag) {
			suplogdebug("%s = (i64)%s", skey, val);
			res = mos_strto64(val, 0, &i64);
			if (res != 0) {
				suplogerr("invalid integer: %s", val);
				goto bad;
			}
			res = pconf_addi(pc, i64, "%s", skey);
			if (res != EPHIDGET_OK) {
				suplogerr("failed to set key '%s'", skey);
				goto bad;
			}
		}
	} else if (gkey) {
		suplogdebug("getting %s", gkey);
		res = pconf_tostring(pc, buf, BUFSIZE, "%s", gkey);
		if (res != EPHIDGET_OK) {
			suplogerr("failed to get key: '%s': %s", gkey, getErrorStr(res));
			goto bad;
		}
		printf("%s", buf);
		goto done;
	}

	if (outfile) {
		suplogdebug("outfile:%s", outfile);
		if (jflag) {
			suplogdebug("rendering new pconf");
			res = pconf_renderpc(pc, buf, BUFSIZE);
			if (res != EPHIDGET_OK) {
				suplogerr("failed to render pconf");
				goto bad;
			}
		} else {
			suplogdebug("rendering new json");
			res = pconf_renderjson(pc, buf, BUFSIZE);
			if (res != EPHIDGET_OK) {
				suplogerr("failed to render JSON");
				goto bad;
			}
		}
		res = writex(buf, mos_strlen(buf), outfile);
		if (res != EPHIDGET_OK) {
			suplogerr("failed to write '%s': %d", outfile, res);
			goto bad;
		}
	} else {
		suplogdebug("rendering pconf");
		res = pconf_renderpc_locked(&pc);
		if (res != EPHIDGET_OK) {
			suplogerr("failed to render pconf");
			goto bad;
		}
	}

done:
bad:

	if (buf)
		mos_free(buf, BUFSIZE);

	if (pc) {
		if (jflag == 0)
			pconf_unlock_locked(&pc);
		else
			pconf_release(&pc);
	}

	PhidgetLog_disable();
	Phidget_finalize(0);

	exit(res);
}